#   Maybe set up private virtual environment

set( 
    WALBERLA_CODEGEN_PRIVATE_VENV ON
    CACHE BOOL 
    "Create a private virtual Python environment inside the build tree for code generation" 
)

if( WALBERLA_CODEGEN_PRIVATE_VENV )
    find_package( Python COMPONENTS Interpreter REQUIRED )

    set(WALBERLA_CODEGEN_VENV_PATH ${CMAKE_CURRENT_BINARY_DIR}/codegen-venv CACHE PATH "Location of the virtual environment used for code generation")
    set(_venv_python_exe ${WALBERLA_CODEGEN_VENV_PATH}/bin/python)
    set(
        _WALBERLA_CODEGEN_VENV_STATEFILE
        ${CMAKE_CURRENT_BINARY_DIR}/walberla-venv-state.json
        CACHE INTERNAL
        "venv statefile - for internal use only"
    )
    set(
        _WALBERLA_CODEGEN_VENV_INVOKE_MANAGER
            ${Python_EXECUTABLE}
            ${sfg_walberla_SOURCE_DIR}/cmake/ManageCodegenVenv.py
            -s
            ${_WALBERLA_CODEGEN_VENV_STATEFILE}
        CACHE INTERNAL
        "venv manager filepath - for internal use only"
    )

    set(
        WALBERLA_CODEGEN_VENV_REQUIREMENTS
        ${sfg_walberla_SOURCE_DIR}/cmake/codegen-requirements.txt
        CACHE PATH
        "Location of the primary requirements file for the codegen virtual environment"
    )
    mark_as_advanced(WALBERLA_CODEGEN_VENV_REQUIREMENTS)


    set(
        _requirements_file
        ${CMAKE_CURRENT_BINARY_DIR}/codegen-requirements.txt
    )

    file(COPY_FILE ${WALBERLA_CODEGEN_VENV_REQUIREMENTS} ${_requirements_file} )
    file(APPEND ${_requirements_file} "\n-e ${sfg_walberla_SOURCE_DIR}\n")

    set(_init_args ${WALBERLA_CODEGEN_VENV_PATH} ${_requirements_file})

    if(NOT _walberla_codegen_venv_initialized)
        #   Force-rebuild environment if cmake cache was deleted
        list(APPEND _init_args --force)
    endif()

    execute_process(
        COMMAND
            $CACHE{_WALBERLA_CODEGEN_VENV_INVOKE_MANAGER}
            init
            ${_init_args}
        RESULT_VARIABLE _lastResult
        ERROR_VARIABLE _lastError
    )

    if( ${_lastResult} )
        message( FATAL_ERROR "Codegen virtual environment setup failed:\n${_lastError}" )
    endif()

    set(
        _walberla_codegen_venv_initialized
        TRUE
        CACHE INTERNAL "venv initialized - for internal use only"
    )

    set( _wlb_codegen_python_init ${_venv_python_exe} )
else()
    #   Use the external Python environment, but check if all packages are installed
    find_package( Python COMPONENTS Interpreter REQUIRED )

    execute_process(
        COMMAND ${Python_EXECUTABLE} ${CMAKE_CURRENT_SOURCE_DIR}/cmake/check_python_env.py
        RESULT_VARIABLE _package_check_status
        OUTPUT_VARIABLE _package_check_output
        ERROR_VARIABLE _package_check_error
    )

    message( STATUS ${_package_check_output} )

    if(NOT (${_package_check_status} EQUAL 0) )
        message( FATAL_ERROR ${_package_check_error} )
    endif()

    set( _wlb_codegen_python_init ${Python_EXECUTABLE} )
endif()

set(PystencilsSfg_PYTHON_PATH ${_wlb_codegen_python_init})
set(
    WALBERLA_CODEGEN_PYTHON
    ${_wlb_codegen_python_init}
    CACHE PATH
    "Path to the Python interpreter used for code generation"
)
mark_as_advanced(WALBERLA_CODEGEN_PYTHON)

#   Find pystencils-sfg

find_package( PystencilsSfg REQUIRED )

#   Project Configuration Module

set( 
    WALBERLA_CODEGEN_CONFIG_MODULE
    ${CMAKE_BINARY_DIR}/CodegenConfig.py
    CACHE
    FILEPATH
    "Path to waLBerla-wide codegen config module" 
)
mark_as_advanced( WALBERLA_CODEGEN_CONFIG_MODULE )

configure_file(
    ${CMAKE_CURRENT_SOURCE_DIR}/cmake/CodegenConfig.template.py
    ${WALBERLA_CODEGEN_CONFIG_MODULE}
)

message( STATUS "Wrote project-wide code generator configuration to ${WALBERLA_CODEGEN_CONFIG_MODULE}" )

function(walberla_codegen_venv_require)
    if(NOT WALBERLA_CODEGEN_PRIVATE_VENV)
        return()
    endif()

    execute_process(
        COMMAND
            $CACHE{_WALBERLA_CODEGEN_VENV_INVOKE_MANAGER}
            require
            --
            ${ARGV}
        RESULT_VARIABLE _lastResult
        ERROR_VARIABLE _lastError
    )

    if( ${_lastResult} )
        message( FATAL_ERROR ${_lastError} )
    endif()
endfunction()

function(walberla_codegen_venv_populate)
    if(NOT WALBERLA_CODEGEN_PRIVATE_VENV)
        return()
    endif()

    execute_process(
        COMMAND
            $CACHE{_WALBERLA_CODEGEN_VENV_INVOKE_MANAGER}
            populate
        RESULT_VARIABLE _lastResult
        ERROR_VARIABLE _lastError
    )

    if( ${_lastResult} )
        message( FATAL_ERROR ${_lastError} )
    endif()
endfunction()

#   Code Generation Functions

#[[
Register code generation scripts for a CMake target.

Signature:

```
walberla_generate_sources( <target> 
    SCRIPTS script1.py [script2.py ...]
    [DEPENDS dependency1.py [dependency2.py...] ]
    [FILE_EXTENSIONS <header-extension> <impl-extension>]
    [OUTPUT_MODE <standalone|inline|header-only>]
)
```

This is a wrapper around `pystencilssfg_generate_target_sources`
without the `CONFIG_MODULE` parameter.
See also https://pycodegen.pages.i10git.cs.fau.de/pystencils-sfg/usage/project_integration.html#add-generator-scripts
#]]
function(walberla_generate_sources TARGET)
    pystencilssfg_generate_target_sources(${ARGV} CONFIG_MODULE $CACHE{WALBERLA_CODEGEN_CONFIG_MODULE})
endfunction()
