#pragma once

#include "blockforest/StructuredBlockForest.h"

#include "core/DataTypes.h"
#include "core/cell/Cell.h"

#include "domain_decomposition/BlockDataID.h"

#include "field/FlagField.h"

#include "stencil/Directions.h"

#include <memory>
#include <tuple>

#include "walberla/experimental/sweep/SparseIndexList.hpp"

namespace walberla::experimental::lbm
{

struct HbbLink
{
   int32_t x;
   int32_t y;
   int32_t z;
   int32_t dir;

   /**
    * Create a halfway bounce-back boundary link from a fluid cell `c` to a boundary cell in direction `d`.
    */
   HbbLink(const Cell& c, const stencil::Direction d)
      : x{ int32_c(c.x()) }, y{ int32_c(c.y()) }, z{ int32_c(c.z()) }, dir{ int32_t(d) }
   {}

   bool operator==(const HbbLink& other) const
   {
      return std::tie(x, y, z, dir) == std::tie(other.x, other.y, other.z, other.dir);
   }
};

template< typename Impl >
struct GenericHbbFactoryImplTraits
{
   using Stencil_T         = typename Impl::Stencil;
   using IdxStruct_T       = HbbLink;
   using IndexListMemTag_T = typename Impl::memtag_t;
   using IndexList_T       = sweep::SparseIndexList< IdxStruct_T, IndexListMemTag_T >;
};

template< typename Impl >
class GenericHbbFactory
{
   using ImplTraits = GenericHbbFactoryImplTraits< Impl >;

 public:
   GenericHbbFactory(const shared_ptr< StructuredBlockForest > blocks, BlockDataID pdfFieldID)
      : blocks_{ blocks }, pdfFieldID_{ pdfFieldID }
   {}

   template< typename FlagField_T >
   auto fromFlagField(BlockDataID flagFieldID, field::FlagUID boundaryFlagUID, field::FlagUID fluidFlagUID)
   {
      using flag_t = typename FlagField_T::flag_t;
      using IndexList_T = typename ImplTraits::IndexList_T;
      using Stencil = typename ImplTraits::Stencil_T;

      IndexList_T indexList{ *blocks_ };

      for (auto& block : *blocks_)
      {
         FlagField_T* flagField = block.getData< FlagField_T >(flagFieldID);
         flag_t boundaryFlag{ flagField->getFlag(boundaryFlagUID) };
         flag_t fluidFlag{ flagField->getFlag(fluidFlagUID) };

         if (!(flagField->flagExists(boundaryFlagUID) && flagField->flagExists(fluidFlagUID))) continue;

         auto& idxVector = indexList.getVector(block);

         for (auto it = flagField->beginXYZ(); it != flagField->end(); ++it)
         {
            Cell fluidCell{ it.cell() };
            if (!isFlagSet(it, fluidFlag)) { continue; }

            for (auto dIt = Stencil::beginNoCenter(); dIt != Stencil::end(); ++dIt)
            {
               stencil::Direction dir{ *dIt };
               Cell neighbor{ fluidCell + dir };
               if (flagField->isFlagSet(neighbor, boundaryFlag)) { idxVector.emplace_back(fluidCell, dir); }
            }
         }
      }

      return impl().irregularFromIndexVector(indexList);
   }

 protected:
   shared_ptr< StructuredBlockForest > blocks_;
   BlockDataID pdfFieldID_;

 private:
   Impl& impl() { return *static_cast< Impl* >(this); }

   const Impl& impl() const { return *static_cast< const Impl* >(this); }
};

} // namespace walberla::experimental::lbm
