from __future__ import annotations

from dataclasses import dataclass, field

from pystencils import CreateKernelConfig, Target
from pystencils.types.quick import Fp
from pystencils.jit import no_jit

from pystencilssfg import SfgContext
from pystencilssfg.composer import SfgIComposer


def cmake_parse_bool(var: str):
    var = var.upper()
    if var in ("ON", "1", "TRUE"):
        return True
    elif var in ("OFF", "0", "FALSE"):
        return False
    else:
        raise ValueError(f"Could not parse cmake value `{var}` as boolean.")


@dataclass
class CodegenOverrides:
    target: Target | None = None
    """Override code generation target"""


@dataclass
class WalberlaBuildConfig:
    """Represents a waLBerla build system configuration"""

    c_compiler_id: str
    """Value of `CMAKE_C_COMPILER_ID`."""

    cxx_compiler_id: str
    """Value of `CMAKE_CXX_COMPILER_ID`."""

    use_double_precision: bool
    """Value of `WALBERLA_DOUBLE_ACCURACY`"""

    optimize_for_localhost: bool
    """Value of `WALBERLA_OPTIMIZE_FOR_LOCALHOST`."""

    mpi_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_MPI`."""

    openmp_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_OPENMP`."""

    cuda_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_CUDA`."""

    hip_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_HIP`."""

    likwid_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_LIKWID_MARKERS`"""

    override: CodegenOverrides = field(default_factory=CodegenOverrides)
    """Override code generator options that would otherwise be inferred from the build system."""

    @staticmethod
    def from_sfg(sfg: SfgContext | SfgIComposer) -> WalberlaBuildConfig:
        if isinstance(sfg, SfgIComposer):
            ctx = sfg.context
        else:
            ctx = sfg

        if isinstance(ctx.project_info, WalberlaBuildConfig):
            return ctx.project_info
        elif DEBUG_MOCK_CMAKE.BUILD_CONFIG is not None:
            return DEBUG_MOCK_CMAKE.BUILD_CONFIG
        else:
            raise ValueError(
                "The given SfgContext does not encapsulate a waLBerla build config object."
            )

    def get_pystencils_config(self) -> CreateKernelConfig:
        cfg = CreateKernelConfig()
        cfg.default_dtype = Fp(64) if self.use_double_precision else Fp(32)
        cfg.jit = no_jit

        if self.override.target is not None:
            cfg.target = self.override.target
        elif self.cuda_enabled:
            cfg.target = Target.CUDA
        elif self.hip_enabled:
            cfg.target = Target.HIP
        else:
            #  CPU target
            if self.optimize_for_localhost:
                cfg.target = Target.CurrentCPU
            else:
                cfg.target = Target.GenericCPU

        if self.openmp_enabled:
            cfg.cpu.openmp.enable = True

        return cfg


def get_build_config(sfg: SfgContext | SfgIComposer):
    """Get the waLBerla build config object for the current generator script."""
    return WalberlaBuildConfig.from_sfg(sfg)


class DEBUG_MOCK_CMAKE:
    BUILD_CONFIG: WalberlaBuildConfig | None = None

    @staticmethod
    def use_cpu_default():
        DEBUG_MOCK_CMAKE.BUILD_CONFIG = WalberlaBuildConfig(
            c_compiler_id="GNU",
            cxx_compiler_id="GNU",
            use_double_precision=True,
            optimize_for_localhost=False,
            mpi_enabled=False,
            openmp_enabled=False,
            hip_enabled=False,
            cuda_enabled=False,
            likwid_enabled=False,
        )

    @staticmethod
    def use_hip_default():
        DEBUG_MOCK_CMAKE.BUILD_CONFIG = WalberlaBuildConfig(
            c_compiler_id="hipcc",
            cxx_compiler_id="hipcc",
            use_double_precision=True,
            optimize_for_localhost=False,
            mpi_enabled=False,
            openmp_enabled=False,
            hip_enabled=True,
            cuda_enabled=False,
            likwid_enabled=False,
        )
