from __future__ import annotations

from typing import Sequence, Iterable, cast, Callable
from dataclasses import dataclass
from itertools import chain
from collections import defaultdict
from abc import ABC, abstractmethod
import warnings

import sympy as sp

from pystencilssfg.composer.custom import CustomGenerator
from pystencilssfg.composer.class_composer import SfgClassComposer, SfgMethodSequencer

from pystencils import (
    Assignment,
    AssignmentCollection,
    CreateKernelConfig,
    Field,
    FieldType,
    Target,
    create_type,
)
from pystencils.types import PsType, constify, deconstify, PsCustomType, PsStructType

from pystencilssfg import SfgComposer
from pystencilssfg.lang import (
    VarLike,
    asvar,
    SfgVar,
    AugExpr,
    Ref,
    SupportsFieldExtraction,
    SupportsVectorExtraction,
)
from pystencilssfg.lang.cpp import std
from pystencilssfg.ir import SfgKernelHandle, SfgCallTreeNode
from .build_config import WalberlaBuildConfig
from .reflection import GeneratedClassWrapperBase
from .api import (
    StructuredBlockForest,
    GenericWalberlaField,
    GhostLayerFieldPtr,
    GpuFieldPtr,
    IndexListBufferPtr,
    SparseIndexList,
    IBlockPtr,
    BlockDataID,
    Vector2,
    Vector3,
    CellInterval,
    CellIdx,
    SharedPtr,
)


class SweepClassProperties:

    @dataclass(frozen=True)
    class Property:
        name: str
        dtype: PsType
        getter: bool
        setter: bool
        initializer: tuple[AugExpr, ...] | None = None

        def __post_init__(self):
            if self.dtype.const and self.setter:
                raise ValueError("Constant properties cannot have setters.")

        @property
        def var(self) -> SfgVar:
            return SfgVar(self.name, self.dtype)

        @property
        def member_var(self) -> SfgVar:
            return SfgVar(self.name + "_", self.dtype)

        def make_methods(self, sfg: SfgComposer) -> Sequence[SfgMethodSequencer]:
            methods = []

            if self.getter:
                methods.append(
                    sfg.method(f"{self.name}")
                    .returns(Ref(constify(self.dtype)))
                    .const()
                    .inline()(f"return {self.name}_;")
                )

            if self.setter:
                methods.append(
                    sfg.method(f"{self.name}")
                    .returns(Ref(self.dtype))
                    .inline()(f"return {self.name}_;")
                )

            return methods

    blockforest_shared_ptr = StructuredBlockForest.shared_ptr()

    def __init__(self) -> None:
        self._properties: dict[str, SweepClassProperties.Property] = dict()

    @property
    def properties(self) -> Iterable[Property]:
        return self._properties.values()

    def add_property(
        self,
        prop: VarLike,
        const=False,
        setter: bool = True,
        getter: bool = True,
        initializer: tuple[AugExpr, ...] | None = None,
    ):
        prop = asvar(prop)
        if prop.name in self._properties:
            raise ValueError(f"Duplicate property: {prop}")

        dtype = constify(prop.dtype) if const else deconstify(prop.dtype)

        self._properties[prop.name] = SweepClassProperties.Property(
            prop.name, dtype, getter, setter, initializer
        )

    def add_blockforest_shared_ptr(self) -> SharedPtr:
        self.add_property(
            self.blockforest_shared_ptr,
            getter=False,
            setter=False,
            initializer=(StructuredBlockForest.shared_ptr_ref(),),
        )
        return self.blockforest_shared_ptr

    def get(self, prop: VarLike) -> AugExpr:
        prop = asvar(prop)

        if prop.name not in self._properties:
            raise KeyError(f"Property not registered: {prop}")

        return AugExpr.format("this->{}_", prop.name)

    def render(self, sfg: SfgComposer):
        priv = sfg.private(*(p.member_var for p in self.properties))

        ctor = self._constructor(sfg)

        for p in self.properties:
            if p.initializer is None:
                ctor.add_param(p.var)
            else:
                for var in chain.from_iterable(e.depends for e in p.initializer):
                    if var not in ctor.parameters:
                        if var.name == asvar(self.blockforest_shared_ptr).name:
                            ctor.add_param(var, 0)
                        else:
                            ctor.add_param(var)

        for p in self._properties.values():
            if p.initializer is not None:
                ctor.init(p.member_var)(*p.initializer)
            else:
                ctor.init(p.member_var)(p.var)

        methods = chain.from_iterable(p.make_methods(sfg) for p in self.properties)
        pub = sfg.public(ctor, *(methods))

        return priv, pub

    def _constructor(self, sfg: SfgComposer) -> SfgClassComposer.ConstructorBuilder:
        return sfg.constructor()


class BlockforestParameters:
    def __init__(
        self,
        property_cache: SweepClassProperties,
        block: IBlockPtr,
    ):
        self._property_cache = property_cache
        self._block = block
        self._extractions: dict[SfgVar, Callable[[CellInterval | None], AugExpr]] = (
            dict()
        )

        self._blockforest: StructuredBlockForest | None = None

    @staticmethod
    def process(asms: AssignmentCollection) -> AssignmentCollection:
        from sympy.core.function import AppliedUndef

        expandable_appls: set[AppliedUndef] = filter(  # type: ignore
            lambda expr: hasattr(expr, "expansion_func"), asms.atoms(AppliedUndef)
        )

        subs: dict[AppliedUndef, sp.Symbol] = dict()
        for appl in expandable_appls:
            expansion: sp.Expr = appl.expansion_func(*appl.args)  # type: ignore
            symb = next(asms.subexpression_symbol_generator)  # type: ignore
            asms.subexpressions.insert(0, Assignment(symb, expansion))
            subs[appl] = symb

        return asms.new_with_substitutions(subs)

    def blockforest(self) -> StructuredBlockForest:
        if self._blockforest is None:
            blocks = self._property_cache.add_blockforest_shared_ptr()
            self._blockforest = StructuredBlockForest().bind(
                "(*{})", self._property_cache.get(blocks)
            )
            return self._blockforest
        else:
            return self._blockforest

    def filter_params(self, params: set[SfgVar]) -> set[SfgVar]:
        from .symbolic import domain, domain_cell_bb, block, block_cell_bb, cell

        params_filtered = set()

        def get_mapping(param: SfgVar, coord: int):
            #   We're populating the mappings with lambdas here,
            #   to make sure that `self.blockforest()` is only called
            #   iff it is actually reqired.
            #   Otherwise blockforest will appear in the parameter list
            #   of sweeps which don't actually need it.
            return {
                domain.aabb_min[coord].name: lambda _: (
                    self.blockforest().getDomain().min()[coord]
                ),
                domain.aabb_max[coord].name: lambda _: (
                    self.blockforest().getDomain().max()[coord]
                ),
                domain_cell_bb.cell_bb_min[coord].name: lambda _: (
                    self.blockforest()
                    .getDomainCellBB(self.blockforest().getLevel(self._block.deref()))
                    .min()[coord]
                ),
                domain_cell_bb.cell_bb_max[coord].name: lambda _: (
                    self.blockforest()
                    .getDomainCellBB(self.blockforest().getLevel(self._block.deref()))
                    .max()[coord]
                ),
                block.aabb_min[coord].name: lambda _: (
                    self._block.getAABB().min()[coord]
                ),
                block.aabb_max[coord].name: lambda _: (
                    self._block.getAABB().max()[coord]
                ),
                block.ci_min[coord].name: lambda ci: (
                    AugExpr.format("0") if ci is None else ci.min()[coord]
                ),
                block.ci_max[coord].name: lambda ci: (
                    AugExpr.format("0") if ci is None else ci.max()[coord]
                ),
                block_cell_bb.cell_bb_min[coord].name: lambda _: (
                    self.blockforest().getBlockCellBB(self._block.deref()).min()[coord]
                ),
                block_cell_bb.cell_bb_max[coord].name: lambda _: (
                    self.blockforest().getBlockCellBB(self._block.deref()).max()[coord]
                ),
                cell.cell_extents[coord].name: lambda _: (
                    self.blockforest().cell_extents(
                        coord, self.blockforest().getLevel(self._block.deref())
                    )
                ),
            }.get(param.name, None)

        for param in params:
            for coord in range(3):
                if (extraction := get_mapping(param, coord)) is not None:
                    self._extractions[param] = extraction
                    break
            else:
                params_filtered.add(param)

        return params_filtered

    def render_extractions(self, sfg: SfgComposer, ci: CellInterval | None):
        return (sfg.init(p)(expr(ci)) for p, expr in self._extractions.items())


@dataclass
class FieldInfo(ABC):
    field: Field

    @property
    @abstractmethod
    def entity(self) -> AugExpr: ...

    @property
    @abstractmethod
    def view(self) -> AugExpr: ...

    @abstractmethod
    def create_view(self, block: IBlockPtr) -> AugExpr: ...

    def view_extraction(self) -> SupportsFieldExtraction:
        assert isinstance(self.view, SupportsFieldExtraction)
        return self.view


@dataclass
class GlFieldInfo(FieldInfo):
    glfield: GenericWalberlaField
    data_id: BlockDataID

    @property
    def entity(self) -> AugExpr:
        return self.data_id

    @property
    def view(self) -> AugExpr:
        return self.glfield

    def create_view(self, block: IBlockPtr) -> AugExpr:
        return block.getData(self.glfield.field_type, self.data_id)


@dataclass
class IndexListInfo(FieldInfo):
    idx_list: SparseIndexList
    idx_vector_view: std.span

    @property
    def entity(self) -> AugExpr:
        return self.idx_list

    @property
    def view(self) -> AugExpr:
        return self.idx_vector_view

    def create_view(self, block: IBlockPtr) -> AugExpr:
        return self.idx_list.view(block.deref())


class ShadowFieldCache:
    """
    Information needed:
        - Name and data type of field
        - Name of shadow field
    """

    def __init__(self) -> None:
        self._originals: dict[str, GlFieldInfo] = dict()

    def add_shadow(self, original: GlFieldInfo):
        self._originals[original.field.name] = original

    def get_shadow(self, original_name: str) -> AugExpr:
        original = self._originals[original_name]
        return AugExpr.format(
            "this->{}({})", self._getter(original.field.name), original.glfield
        )

    def perform_swap(self, original_name: str, shadow_field: GlFieldInfo) -> AugExpr:
        original = self._originals[original_name]
        return original.glfield.swapDataPointers(shadow_field.glfield)

    def render(self, sfg: SfgComposer):
        if not self._originals:
            return ()

        sfg.include("<memory>")

        cache_ptrs = []
        getters = []

        for orig_name, orig in self._originals.items():
            unique_ptr_type = PsCustomType(
                f"std::unique_ptr< {orig.glfield.field_type} >"
            )
            cache_ptr_name = self._cache_ptr(orig.glfield)
            cache_ptrs.append(sfg.var(cache_ptr_name, unique_ptr_type))

            getters.append(
                sfg.method(self._getter(orig_name))
                .returns(orig.glfield.get_dtype())
                .inline()(
                    sfg.branch(f"{cache_ptr_name} == nullptr")(
                        AugExpr.format(
                            "{}.reset({});",
                            cache_ptr_name,
                            orig.glfield.cloneUninitialized(),
                        )
                    ),
                    f"return {cache_ptr_name}.get();",
                )
            )

        return (sfg.private(*cache_ptrs, *getters),)

    def _getter(self, orig_name: str) -> str:
        return f"_getShadow_{orig_name}"

    def _cache_ptr(self, orig: GenericWalberlaField) -> str:
        return f"shadow_{str(orig)}_"


def combine_vectors(
    scalars: set[SfgVar],
) -> dict[SupportsVectorExtraction, tuple[SfgVar, ...]]:
    """Attempt to combine vector component symbols into vectors.

    This function modifies the `scalars` parameter in-place by removing grouped scalar
    components and replacing them by their vectors.
    """
    potential_vectors: defaultdict[str, list[tuple[int, SfgVar]]] = defaultdict(list)

    for s in scalars:
        tokens = s.name.split("_")
        try:
            coord = int(tokens[-1])
            vec_name = "_".join(tokens[:-1])
            potential_vectors[vec_name].append((coord, s))
        except ValueError:
            pass

    all_vectors: dict[SupportsVectorExtraction, tuple[SfgVar, ...]] = dict()

    for name, entries in potential_vectors.items():
        entries = sorted(entries, key=lambda t: t[0])
        coords = [e[0] for e in entries]
        components = tuple(e[1] for e in entries)
        scalar_types = set(e[1].dtype for e in entries)

        vec_type: type[AugExpr]
        if set(coords) == {0, 1}:
            vec_type = Vector2
        elif set(coords) == {0, 1, 2}:
            vec_type = Vector3
        else:
            continue

        if len(scalar_types) > 1:
            warnings.warn(
                f"Skipping vector grouping for vector {name}: "
                f"Encountered vector components {components} with conflicting data types {scalar_types}.",
                UserWarning,
            )
            continue

        name_conflict = False
        for s in scalars:
            if s.name == name:
                warnings.warn(
                    f"Skipping vector grouping for vector {name}: "
                    f"Vector name conflicts with existing scalar parameter {s}.",
                    UserWarning,
                )
                name_conflict = True

        if name_conflict:
            continue

        scalar_type = deconstify(scalar_types.pop())
        vector = vec_type(scalar_type).var(name)
        all_vectors[vector] = components

        for c in components:
            scalars.remove(c)
        scalars.add(asvar(vector))

    return all_vectors


class Sweep(CustomGenerator):
    """Generate a waLBerla sweep from a pystencils kernel."""

    def __init__(
        self,
        name: str,
        assignments: Sequence[Assignment] | AssignmentCollection,
        config: CreateKernelConfig | None = None,
    ):
        if config is not None:
            cfg = config.copy()

            if cfg.get_option("ghost_layers") is not None:
                raise ValueError(
                    "Specifying `ghost_layers` in your codegen config is invalid when generating a waLBerla sweep."
                )

            if (
                cfg.get_option("iteration_slice") is None
                and cfg.get_option("index_field") is None
            ):
                cfg.ghost_layers = 0
        else:
            cfg = CreateKernelConfig(ghost_layers=0)

        manual_grid: bool = cfg.gpu.get_option("manual_launch_grid")
        if manual_grid:
            raise ValueError(
                "Setting `gpu.manual_launch_grid = True` is invalid for waLBerla sweeps."
            )

        self._name = name
        self._gen_config = cfg

        if isinstance(assignments, AssignmentCollection):
            self._assignments = assignments
        else:
            self._assignments = AssignmentCollection(assignments)  # type: ignore

        #   Set only later once the full codegen config is known
        self._glfield_type: type[GpuFieldPtr] | type[GhostLayerFieldPtr]

        #   Map from shadow field to shadowed field
        self._shadow_fields: dict[Field, Field] = dict()

        #   RESULTS - unset at this point
        self._generated_class: type[GeneratedClassWrapperBase] | None = None

    #   CONFIGURATION

    @property
    def sparse(self) -> bool:
        return self._gen_config.get_option("index_field") is not None

    @sparse.setter
    def sparse(self, sparse_iteration: bool):
        if sparse_iteration:
            if self._gen_config.get_option("index_field") is not None:
                return

            if self._gen_config.get_option("index_dtype") != create_type("int64"):
                raise ValueError(
                    "Sparse sweeps require `int64_t` as their index data type. Check your code generator config."
                )

            self._gen_config.index_field = Field.create_generic(
                "indexList", 1, CellIdx, field_type=FieldType.INDEXED
            )
            self._gen_config.ghost_layers = None
        else:
            self._gen_config.index_field = None
            self._gen_config.ghost_layers = 0

    def swap_fields(self, field: Field, shadow_field: Field):
        if field in self._shadow_fields:
            raise ValueError(f"Field swap for {field} was already registered.")
        if shadow_field in self._shadow_fields:
            raise ValueError(f"Field swap for {shadow_field} was already registered.")
        if field.dtype != shadow_field.dtype:
            raise ValueError(
                "Field and its shadow must have the same element type for swapping"
            )

        self._shadow_fields[shadow_field] = field

    #   CODE GENERATION

    def _set_field_interface(self, target: Target):
        if target.is_gpu():
            self._glfield_type = GpuFieldPtr
        elif target.is_cpu():
            self._glfield_type = GhostLayerFieldPtr
        else:
            raise ValueError(
                f"Cannot generate sweep for target {self._gen_config.target}"
            )

    def _walberla_field(self, f: Field) -> GenericWalberlaField | IndexListBufferPtr:
        match f.field_type:
            case FieldType.GENERIC | FieldType.CUSTOM:
                return self._glfield_type.create(f)
            case FieldType.INDEXED:
                assert isinstance(f.dtype, PsStructType)
                return IndexListBufferPtr(f.dtype).var(f.name)
            case _:
                raise ValueError(
                    f"Unable to map field {f} of type {f.field_type} to a waLBerla field."
                )

    def _make_field_info(self, f: Field, target: Target) -> FieldInfo:
        match f.field_type:
            case FieldType.GENERIC | FieldType.CUSTOM:
                glfield = self._glfield_type.create(f)
                data_id = BlockDataID().var(f"{f.name}Id")
                return GlFieldInfo(f, glfield, data_id)
            case FieldType.INDEXED:
                assert isinstance(f.dtype, PsStructType)
                idx_list = SparseIndexList.from_field(f, target=target)
                view = idx_list.view_type().var(f"{f.name}_view")
                return IndexListInfo(f, idx_list, view)
            case _:
                raise ValueError(f"Unexpected field type: {f.field_type} at field  {f}")

    def _render_invocation(
        self, sfg: SfgComposer, target: Target, khandle: SfgKernelHandle
    ) -> tuple[SfgCallTreeNode, set[SfgVar]]:
        """Render and return the kernel invocation plus a set of additional parameters required
        at the call site."""

        if target.is_gpu():
            # from pystencils.codegen.config import GpuIndexingScheme

            #   TODO: Want default values for properties first,
            #   to define default stream and block size values
            # indexing_scheme = self._gen_config.gpu.get_option("indexing_scheme")
            # if indexing_scheme == GpuIndexingScheme.Linear3D:
            #     block_size = sfg.gpu_api.dim3(const=True).var("gpuBlockSize")
            #     return (sfg.gpu_invoke(khandle, block_size=block_size), {block_size})
            # else:
            return (sfg.gpu_invoke(khandle), set())

        else:
            return (sfg.call(khandle), set())

    def generate(self, sfg: SfgComposer) -> None:
        build_config = WalberlaBuildConfig.from_sfg(sfg)
        gen_config = build_config.get_pystencils_config()
        gen_config.override(self._gen_config)

        target = gen_config.get_target()
        self._set_field_interface(target)

        knamespace = sfg.kernel_namespace(f"{self._name}_kernels")

        assignments = BlockforestParameters.process(self._assignments)

        khandle = knamespace.create(assignments, self._name, gen_config)
        ker_invocation, ker_call_site_params = self._render_invocation(
            sfg, target, khandle
        )

        all_fields: dict[str, FieldInfo] = {
            f.name: self._make_field_info(f, target) for f in khandle.fields
        }

        swaps: dict[str, GlFieldInfo] = dict()
        shadows_cache = ShadowFieldCache()

        for shadow, orig in self._shadow_fields.items():
            shadow_fi = all_fields[shadow.name]
            assert isinstance(shadow_fi, GlFieldInfo)
            original_fi = all_fields[orig.name]
            assert isinstance(original_fi, GlFieldInfo)
            swaps[orig.name] = shadow_fi
            shadows_cache.add_shadow(original_fi)

        block = IBlockPtr().var("block")
        block_fields = sorted(
            (fi for fi in all_fields.values() if fi.field not in self._shadow_fields),
            key=lambda fi: fi.field.name,
        )

        props = SweepClassProperties()

        parameters = khandle.scalar_parameters | ker_call_site_params

        blockforest_params = BlockforestParameters(props, block)
        parameters = blockforest_params.filter_params(parameters)

        vector_groups = combine_vectors(parameters)

        for fi in block_fields:
            props.add_property(fi.entity, setter=False, getter=True)

        for s in sorted(parameters, key=lambda p: p.name):
            props.add_property(s, setter=True, getter=True)

        def render_runmethod(ci: CellInterval | None = None):
            return [
                #   Get IDs from class
                *(sfg.init(fi.entity)(props.get(fi.entity)) for fi in block_fields),
                #   Extract field views
                *(sfg.init(fi.view)(fi.create_view(block)) for fi in block_fields),
                #   Get shadow fields from cache
                *(
                    sfg.init(shadow_info.glfield)(shadows_cache.get_shadow(orig_name))
                    for orig_name, shadow_info in swaps.items()
                ),
                #   Map GhostLayerFields to pystencils fields
                *(
                    sfg.map_field(fi.field, fi.glfield.with_cell_interval(ci))
                    for fi in all_fields.values()
                    if isinstance(fi, GlFieldInfo)
                ),
                #   Extract indexing information from field view for all remaining fields
                *(
                    sfg.map_field(fi.field, fi.view_extraction())
                    for fi in all_fields.values()
                    if not isinstance(fi, GlFieldInfo)
                ),
                #   Get parameters from class
                *(sfg.init(param)(props.get(param)) for param in parameters),
                #   Extract scalars from vectors
                *(
                    sfg.map_vector(components, vector)
                    for vector, components in vector_groups.items()
                ),
                #   Extract geometry information
                *(blockforest_params.render_extractions(sfg, ci)),
                #   Invoke the kernel
                ker_invocation,
                #   Perform field swaps
                *(
                    shadows_cache.perform_swap(orig_name, shadow_info)
                    for orig_name, shadow_info in swaps.items()
                ),
            ]

        methods = [sfg.method("operator()")(*render_runmethod())]

        if not self.sparse:
            methods.append(
                sfg.method("runOnCellInterval")(
                    *render_runmethod(CellInterval(const=True, ref=True).var("ci"))
                )
            )

        sfg.klass(self._name)(
            *props.render(sfg),
            sfg.public(*methods),
            *shadows_cache.render(sfg),
        )

        gen_class = sfg._cursor.get_entity(self._name)
        assert gen_class is not None
        from pystencilssfg.ir.entities import SfgClass

        assert isinstance(gen_class, SfgClass)

        class GenClassWrapper(GeneratedClassWrapperBase):
            _class = cast(SfgClass, gen_class)

        self._generated_class = GenClassWrapper

    #   CODE-GENERATION RESULTS
    #   These properties are only available after the sweep was generated

    @property
    def generated_class(self) -> type[GeneratedClassWrapperBase]:
        if self._generated_class is None:
            raise AttributeError(
                "Generated class is unavailable - code generation was not run yet."
            )

        return self._generated_class
