#include "blockforest/all.h"
#include "blockforest/communication/UniformBufferedScheme.h"

#include "core/all.h"

#include "domain_decomposition/SharedSweep.h"

#include "field/all.h"
#include "field/communication/PackInfo.h"
#include "field/communication/StencilRestrictedPackInfo.h"

#include "geometry/InitBoundaryHandling.h"

#include "stencil/all.h"

#include "timeloop/all.h"

#include "vtk/all.h"

#include "gen/LbmAlgorithms.hpp"

namespace Ex_ForceDrivenChannel
{
    using namespace walberla;
    using namespace blockforest;

    using std::make_unique;
    using std::shared_ptr;

    using ScalarField_T = field::GhostLayerField<real_t, 1>;
    using VectorField_T = field::GhostLayerField<real_t, 3>;

    using LbStencil = stencil::D3Q19;
    using PdfField_T = field::GhostLayerField<real_t, LbStencil::Q>;

    using CommScheme = blockforest::communication::UniformBufferedScheme<LbStencil>;
    using PdfsPackInfo = field::communication::StencilRestrictedPackInfo<PdfField_T, LbStencil>;

    using FlagField_T = FlagField<uint8_t>;

    void run(const shared_ptr<Config> &config)
    {
        auto blocks = createUniformBlockGridFromConfig(config);

        Config::BlockHandle simParams = config->getBlock("Parameters");
        const real_t omega{simParams.getParameter<real_t>("omega")};

        const Vector3<real_t> force = simParams.getParameter<Vector3<real_t>>("force");

        BlockDataID pdfsId = field::addToStorage<PdfField_T>(blocks, "pdfs", real_c(0.0), field::fzyx, 1);
        BlockDataID rhoId = field::addToStorage<ScalarField_T>(blocks, "rho", real_c(1.0), field::fzyx, 0);
        BlockDataID uId = field::addToStorage<VectorField_T>(blocks, "u", real_c(0.0), field::fzyx, 0);

        gen::LbInit lbInit{pdfsId, rhoId, uId, force};

        for (auto &block : *blocks)
        {
            lbInit(&block);
        }

        //  Stream-Collide

        auto streamCollide = make_shared<gen::LbStreamCollide>(pdfsId, rhoId, uId, force, omega);

        //  Communication Setup

        CommScheme comm{blocks};
        auto pdfsPackInfo = std::make_shared<PdfsPackInfo>(pdfsId);
        comm.addPackInfo(pdfsPackInfo);

        //  Boundary Conditions

        const BlockDataID flagFieldId = field::addFlagFieldToStorage<FlagField_T>(blocks, "flagField");
        const FlagUID fluidFlagUid{"Fluid"};
        const FlagUID noSlipFlagUid{"NoSlip"};

        auto boundariesConfig = config->getBlock("Boundaries");
        if (boundariesConfig)
        {
            WALBERLA_LOG_INFO_ON_ROOT("Setting boundary conditions")
            geometry::initBoundaryHandling<FlagField_T>(*blocks, flagFieldId, boundariesConfig);
        }

        geometry::setNonBoundaryCellsToDomain<FlagField_T>(*blocks, flagFieldId, fluidFlagUid);

        auto flagFieldOutput = field::createVTKOutput<FlagField_T>(flagFieldId, *blocks, "flagField", 1, 0);
        flagFieldOutput();

        gen::NoSlip noSlip {
            gen::NoSlipFactory(blocks, pdfsId).fromFlagField< FlagField_T >(flagFieldId, noSlipFlagUid, fluidFlagUid)
        };

        auto boundarySweep = [&](IBlock *block)
        {
            noSlip(block);
        };

        //  Timeloop
        const uint_t numTimesteps{simParams.getParameter<uint_t>("timesteps")};
        SweepTimeloop loop{blocks->getBlockStorage(), numTimesteps};

        loop.add() << Sweep(makeSharedSweep(streamCollide));
        loop.add() << Sweep(boundarySweep) << AfterFunction(comm);

        RemainingTimeLogger logger{numTimesteps};
        loop.addFuncAfterTimeStep(logger);

        //  VTK Output

        Config::BlockHandle outputParams = config->getBlock("Output");

        const uint_t vtkWriteFrequency = outputParams.getParameter<uint_t>("vtkWriteFrequency", 0);
        if (vtkWriteFrequency > 0)
        {
            auto vtkOutput = vtk::createVTKOutput_BlockData(*blocks, "vtk", vtkWriteFrequency, 0, false, "vtk_out",
                                                            "simulation_step", false, true, true, false, 0);

            auto densityWriter = make_shared<field::VTKWriter<ScalarField_T, float32>>(rhoId, "density");
            vtkOutput->addCellDataWriter(densityWriter);

            auto velWriter = make_shared<field::VTKWriter<VectorField_T, float32>>(uId, "velocity");
            vtkOutput->addCellDataWriter(velWriter);

            const cell_idx_t xCells = cell_idx_c(blocks->getNumberOfXCells());
            const cell_idx_t yCells = cell_idx_c(blocks->getNumberOfYCells());
            const cell_idx_t zCells = cell_idx_c(blocks->getNumberOfZCells());

            loop.addFuncAfterTimeStep(vtk::writeFiles(vtkOutput), "VTK Output");
        }

        //  Run the Simulation

        WALBERLA_LOG_INFO_ON_ROOT("Commencing simulation with " << numTimesteps << " timesteps")

        loop.run();
    }
}

int main(int argc, char **argv)
{
    walberla::Environment env{argc, argv};

    Ex_ForceDrivenChannel::run(env.config());

    return EXIT_SUCCESS;
}
