# Getting Started with Generator Scripts

This chapter aims to give an introduction on working with *generator scripts*,
which are the driving force of the next-codegen system in waLBerla.

In the course of this guide, we will
 - set up the basic structure of a generator script;
 - register that script with CMake;
 - and include the generated files into a simple *Hello World*-like application.

This guide assumes that you already have a CMake project and Python environment correctly
set up, as described in [](EnvSetup).

:::{dropdown} Get The Files

{download}`GeneratorScriptBasics.zip </zipped-examples/GeneratorScriptBasics.zip>`.

This example comprises the following files:

 - `GeneratorScriptBasics.cpp`: The demo simulation;
 - `BasicCodegen.py`: A sample code generation script;
 - `CMakeLists.txt`: The CMake build system configuration.

:::

## Our First Generator Script

Create and open the file `BasicCodegen.py` and populate it with the following code:

```{literalinclude} BasicCodegen.py
:caption: BasicCodegen.py
:lineno-start: 1
```

Let's take this apart.
- In line 1, we import the `SourceFileGenerator`, which is the object responsible for running code generation
  and interacting with the build system.
  It is exposed by [pystencils-sfg][pystencils-sfg], a lower-level package that `sfg-walberla` relies on.
- In line 3, we enter code generation mode by opening up a managed block controlled by the `SourceFileGenerator`.
  It gives us the `sfg` object, which is our primary interaction point with the code generation engine.
- In lines 4 to 6, we use the `sfg` object to do three things:
    - Include a header file; this will cause the respective `#include` directive to be generated.
    - Set the namespace; this will place all generated code into the `gen` namespace.
    - Define a constant; this will print the definition verbatim into the output header file.

To see the generator script in effect, we need to create a CMake target and a simple application we can include it into.
Create both the `BasicCodegenApp.cpp` and `CMakeLists.txt` files, open up `CMakeLists.txt`, and add the following:

:::{literalinclude} CMakeLists.txt
:language: CMake
:caption: CMakeLists.txt
:lineno-start: 1
:emphasize-lines: 4-6
:::

At first, we create a new executable target called `Ex_GeneratorScriptBasics`
and add the `BasicCodegen.cpp` source file to it.
To us code generators, however, the second part is more interesting:
we register our generator script `BasicCodegen.py` at the target using
`walberla_generate_sources`.
This will cause the build system to execute the generator script during build,
compile its generated sources,
and make its generated headers available to us.

To observe this, add the following to your `BasicCodegenApp.cpp`:

```{literalinclude} BasicCodegenApp.cpp
:caption: BasicCodegenApp.cpp
:lineno-start: 1
```

Don't worry if your IDE tells you `BasicCodegen.hpp` does not exist yet;
the file will be generated in a moment.
Use CMake to build the application `Ex_GeneratorScriptBasics` and run it,
and you should see it print `51966` to stdout.

Let's take a look at what is happening during the build.
Your CMake build output should contain a line somewhat like this:

```
Generating sfg_sources/gen/Ex_GeneratorScriptBasics/BasicCodegen.hpp, sfg_sources/gen/Ex_GeneratorScriptBasics/BasicCodegen.cpp
```

This indicates that the generator script was executed and produced exactly two files below the `sfg_sources` directory:
- a header file `gen/BasicCodegen.hpp`, which is the one we included into our application, and
- a source file `gen/BasicCodegen.cpp`, which at this time is still empty
  because we haven't defined any kernels or functions yet.

The `walberla_generate_sources` CMake command made sure that the `sfg_sources` directory was placed in your
target's include path, so you can include the generated files.

That covers the basics of setting up and running a code generator script;
you're all set now to generate and run your first numerical kernels.

[pystencils-sfg]: https://pycodegen.pages.i10git.cs.fau.de/pystencils-sfg/index.html
[sfg_add_gen_scripts]: https://pycodegen.pages.i10git.cs.fau.de/pystencils-sfg/usage/project_integration.html#add-generator-scripts "pystencils-sfg Documentation"
