#include "core/all.h"
#include "blockforest/all.h"
#include "field/GhostLayerField.h"
#include "field/AddToStorage.h"
#include "field/vtk/all.h"
#include "walberla/experimental/sweep/SparseIndexList.hpp"

#include <array>

#include "gen/SpiralSweep.hpp"

namespace SparseSpiral
{
    using namespace walberla;
    using namespace walberla::experimental;

    using std::array;
    using std::vector;

    using PointField = field::GhostLayerField<real_t, 3>;

    constexpr real_t radius { 15.0 };
    constexpr Vector3< real_t > center { 20, 20, 0 };
    constexpr real_t height { 120. };
    constexpr real_t width { 4.0 };

    Vector3<real_t> pointOnSpiral(real_t z)
    {
        real_t t{4.0 * math::pi * z / height};
        return center + Vector3<real_t>{radius * cos(t), radius * sin(t), z};
    }

    void main(int argc, char **argv)
    {
        Environment env{ argc, argv };

        AABB domainAabb{0., 0., 0., 40., 40., height };
        array<uint_t, 3> gridSize{4, 4, 12};
        array<uint_t, 3> cpb{10, 10, 10};

        auto blocks = blockforest::createUniformBlockGrid(
            domainAabb,
            gridSize[0], gridSize[1], gridSize[2],
            cpb[0], cpb[1], cpb[2]);

        BlockDataID fID = field::addToStorage<PointField>(blocks, "f", real_c(0.0), field::fzyx);
        sweep::SparseIndexList indexList{*blocks};

        auto isOnSpiral = [&](Cell globalCell) -> bool
        {
            Vector3<real_t> globalPoint{blocks->getCellCenter(globalCell)};
            Vector3<real_t> spiralPoint{pointOnSpiral(globalPoint[2])};
            return (globalPoint - spiralPoint).sqrLength() < width;
        };

        CellInterval allCells{{0, 0, 0}, {cell_idx_c(cpb[0]), cell_idx_c(cpb[1]), cell_idx_c(cpb[2])}};

        for (auto &b : *blocks)
        {
            auto & idxVec{indexList.getVector(b)};
            for (Cell c : allCells)
            {
                Cell globalCell{c};
                blocks->transformBlockLocalToGlobalCell(globalCell, b);
                if (isOnSpiral(globalCell))
                {
                    idxVec.emplace_back(c);
                }
            }
        }

        gen::SpiralSweep sweep{blocks, fID, indexList};

        for (auto &b : *blocks)
        {
            sweep(&b);
        }

        for (auto &b : *blocks)
        {
            PointField & f = *b.getData<PointField>(fID);
            for (Cell c : allCells)
            {
                Cell globalCell{c};
                blocks->transformBlockLocalToGlobalCell(globalCell, b);
                if (isOnSpiral(c))
                {
                    Vector3< real_t > cellCenter { blocks->getCellCenter(globalCell) };
                    WALBERLA_CHECK_FLOAT_EQUAL(f.get(c, 0), cellCenter[0]);
                    WALBERLA_CHECK_FLOAT_EQUAL(f.get(c, 1), cellCenter[1]);
                    WALBERLA_CHECK_FLOAT_EQUAL(f.get(c, 2), cellCenter[2]);
                }
            }
        }

        auto vtkWriter = field::createVTKOutput< PointField >(fID, *blocks, "f", 1, 0);
        vtkWriter();
    }
}

int main(int argc, char **argv)
{
    SparseSpiral::main(argc, argv);
    return 0;
}
